#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/cpumask.h>
#include <linux/debugfs.h>
#include <linux/spinlock.h>
#include <linux/module.h>
#include <linux/cpu.h>
#include <linux/dca.h>

static struct dentry *	dca_debug_dir;
static struct dentry *	dca_debug_tag_dir;

#define DCA_TAG_NAME_LEN	7

struct dca_debug_tag {
	struct dentry *		dentry;
	char			name[DCA_TAG_NAME_LEN+1]; /* ensure \0 */
};

static DEFINE_PER_CPU(struct dca_debug_tag, dca_debug_tags);

static u64 debugfs_dca_tag_get(void *data)
{
	int cpu = (unsigned long)data;
	return dca_get_tag(cpu);
}

DEFINE_SIMPLE_ATTRIBUTE(fops_dca_tag, debugfs_dca_tag_get, NULL, "0x%llX\n");

static struct dentry *debugfs_create_dca_tag(const char *name, mode_t mode,
                                             struct dentry *parent, int cpu)
{	
	void *data = (void*)(unsigned long)cpu;
	return debugfs_create_file(name, mode, parent, data, &fops_dca_tag);
}

static int dca_debug_populate_tag(int cpu)
{
	struct dca_debug_tag *entry = &per_cpu(dca_debug_tags, cpu);

	snprintf(entry->name, DCA_TAG_NAME_LEN, "cpu%d", cpu);
	entry->dentry = debugfs_create_dca_tag(entry->name, S_IRUGO,
	                                       dca_debug_tag_dir, cpu);
	if (!entry->dentry)
		return -ENOMEM;
	return 0;
}

static void dca_debug_remove_tag(int cpu)
{
	struct dca_debug_tag *entry = &per_cpu(dca_debug_tags, cpu);
	debugfs_remove(entry->dentry);
}

void dca_debug_remove_all_tags(void)
{
	struct dca_debug_tag *entry;
	int cpu;

	for_each_online_cpu(cpu) {
		entry = &per_cpu(dca_debug_tags, cpu);
		if (entry->dentry)
			debugfs_remove(entry->dentry);
	}
}

int dca_debug_populate_all_tags(void)
{
	int cpu, err;

	for_each_online_cpu(cpu) {
		err = dca_debug_populate_tag(cpu);
		if (err)
			goto undo;
	}
	return 0;
undo:
	dca_debug_remove_all_tags();
	return err;
}

static int dca_debug_cpu_callback(struct notifier_block *nfb, unsigned long action, void *hcpu)
{
	unsigned int cpu = (unsigned long)hcpu;

	switch (action) {
	case CPU_ONLINE:
		dca_debug_populate_tag(cpu);
		break;
	case CPU_DEAD:
		dca_debug_remove_tag(cpu);
		break;
	}
	return NOTIFY_OK;
}

static struct notifier_block __cpuinitdata dca_debug_cpu_notifier = {
	.notifier_call = dca_debug_cpu_callback,
};

int __init dca_debug_init(void)
{
	dca_debug_dir = debugfs_create_dir("dca", NULL);
	if (!dca_debug_dir)
		return -ENOMEM;
	dca_debug_tag_dir = debugfs_create_dir("tags", dca_debug_dir);
	if (!dca_debug_tag_dir) {
		debugfs_remove(dca_debug_dir);
		return -ENOMEM;
	}
	register_hotcpu_notifier(&dca_debug_cpu_notifier);
	return 0;
}

void __exit dca_debug_exit(void)
{
	unregister_hotcpu_notifier(&dca_debug_cpu_notifier);
	debugfs_remove(dca_debug_tag_dir);
	debugfs_remove(dca_debug_dir);
}

